@@ -9,7 +9,6 @@ Functions and general utilities for the deepof package.
 """
 
 import argparse
-import gc
 import multiprocessing
 import os
 from copy import deepcopy
@@ -17,14 +16,17 @@ from itertools import combinations, product
 from typing import Tuple, Any, List, Union, NewType
 
 import cv2
+import math
 import networkx as nx
 import numpy as np
 import pandas as pd
 import regex as re
 import ruptures as rpt
 import tensorflow as tf
+from dask_image.imread import imread
 from joblib import Parallel, delayed
 from scipy.signal import savgol_filter
+from shapely.geometry import Polygon
 from sklearn import mixture
 from sklearn.feature_selection import VarianceThreshold
 from tqdm import tqdm
@@ -159,7 +161,7 @@ def tab2polar(cartesian_df: pd.DataFrame) -> pd.DataFrame:
         result.append(bp2polar(cartesian_df[df]))
     result = pd.concat(result, axis=1)
     idx = pd.MultiIndex.from_product(
-        [list(cartesian_df.columns.levels[0]), ["rho", "phi"]],
+        [list(cartesian_df.columns.levels[0]), ["rho", "phi"]]
     )
     result.columns = idx
     result.index = cartesian_df.index
@@ -265,6 +267,50 @@ def angle_trio(bpart_array: np.array) -> np.array:
     return ang_trio
 
 
+def compute_areas(coords, animal_id=None):
+    """
+    Computes relevant areas (head, torso, back, full) for the provided coordinates.
+    Args:
+        coords: coordinates of the body parts for a single time point.
+        animal_id: animal id for the provided coordinates, if any.
+
+    Returns:
+        areas: list including head, torso, back, and full areas for the provided coordinates.
+
+    """
+
+    head = ["Nose", "Left_ear", "Left_fhip", "Spine_1"]
+
+    torso = ["Spine_1", "Right_fhip", "Spine_2", "Left_fhip"]
+
+    back = ["Spine_1", "Right_bhip", "Spine_2", "Left_bhip"]
+
+    full = [
+        "Nose",
+        "Left_ear",
+        "Left_fhip",
+        "Left_bhip",
+        "Tail_base",
+        "Right_bhip",
+        "Right_fhip",
+        "Right_ear",
+    ]
+
+    areas = []
+
+    for bps in [head, torso, back, full]:
+
+        if animal_id is not None:
+            bps = ["_".join([animal_id, bp]) for bp in bps]
+
+        x = coords.xs(key="x", level=1)[bps]
+        y = coords.xs(key="y", level=1)[bps]
+
+        areas.append(Polygon(zip(x, y)).area)
+
+    return areas
+
+
 def rotate(
     p: np.array, angles: np.array, origin: np.array = np.array([0, 0])
 ) -> np.array:
@@ -292,6 +338,7 @@ def rotate(
     return rotated
 
 
+# noinspection PyArgumentList
 def align_trajectories(data: np.array, mode: str = "all") -> np.array:
     """
 
@@ -327,8 +374,7 @@ def align_trajectories(data: np.array, mode: str = "all") -> np.array:
 
     for frame in range(data.shape[0]):
         aligned_trajs[frame] = rotate(
-            data[frame].reshape([-1, 2], order="C"),
-            angles[frame],
+            data[frame].reshape([-1, 2], order="C"), angles[frame]
         ).reshape(data.shape[1:], order="C")
 
     if mode == "all" or mode == "none":
@@ -337,23 +383,165 @@ def align_trajectories(data: np.array, mode: str = "all") -> np.array:
     return aligned_trajs
 
 
-def smooth_boolean_array(a: np.array) -> np.array:
+def kleinberg(offsets, s=np.e, gamma=1.0, n=None, T=None, k=None):
+    """Kleinberg's algorithm (described in 'Bursty and Hierarchical Structure
+    in Streams'). The algorithm models activity bursts in a time series as an
+    infinite hidden Markov model.
+
+    Taken from pybursts (https://github.com/romain-fontugne/pybursts/blob/master/pybursts/pybursts.py)
+    and adapted for dependency compatibility reasons.
+
+    Input:
+        offsets: a list of time offsets (numeric)
+        s: the base of the exponential distribution that is used for modeling
+        the event frequencies
+        gamma: coefficient for the transition costs between states
+        n, T: to have a fixed cost function (not dependent of the given offsets).
+        Which is needed if you want to compare bursts for different inputs.
+        k: maximum burst level"""
+
+    if s <= 1:
+        raise ValueError("s must be greater than 1!")
+    if gamma <= 0:
+        raise ValueError("gamma must be positive!")
+    if not n is None and n <= 0:
+        raise ValueError("n must be positive!")
+    if not T is None and T <= 0:
+        raise ValueError("T must be positive!")
+    if len(offsets) < 1:
+        raise ValueError("offsets must be non-empty!")
+
+    offsets = np.array(offsets, dtype=object)
+
+    if offsets.size == 1:
+        bursts = np.array([0, offsets[0], offsets[0]], ndmin=2, dtype=object)
+        return bursts
+
+    offsets = np.sort(offsets)
+    gaps = np.diff(offsets)
+
+    if not np.all(gaps):
+        raise ValueError("Input cannot contain events with zero time between!")
+
+    if T is None:
+        T = np.sum(gaps)
+
+    if n is None:
+        n = np.size(gaps)
+
+    g_hat = T / n
+    gamma_log_n = gamma * math.log(n)
+
+    if k is None:
+        k = int(math.ceil(float(1 + math.log(T, s) + math.log(1 / np.amin(gaps), s))))
+
+    def tau(i, j):
+        if i >= j:
+            return 0
+        else:
+            return (j - i) * gamma_log_n
+
+    alpha_function = np.vectorize(lambda x: s ** x / g_hat)
+    alpha = alpha_function(np.arange(k))
+
+    def f(j, x):
+        return alpha[j] * math.exp(-alpha[j] * x)
+
+    C = np.repeat(float("inf"), k)
+    C[0] = 0
+
+    q = np.empty((k, 0))
+    for t in range(np.size(gaps)):
+        C_prime = np.repeat(float("inf"), k)
+        q_prime = np.empty((k, t + 1))
+        q_prime.fill(np.nan)
+
+        for j in range(k):
+            cost_function = np.vectorize(lambda x: C[x] + tau(x, j))
+            cost = cost_function(np.arange(0, k))
+
+            el = np.argmin(cost)
+
+            if f(j, gaps[t]) > 0:
+                C_prime[j] = cost[el] - math.log(f(j, gaps[t]))
+
+            if t > 0:
+                q_prime[j, :t] = q[el, :]
+
+            q_prime[j, t] = j + 1
+
+        C = C_prime
+        q = q_prime
+
+    j = np.argmin(C)
+    q = q[j, :]
+
+    prev_q = 0
+
+    N = 0
+    for t in range(np.size(gaps)):
+        if q[t] > prev_q:
+            N = N + q[t] - prev_q
+        prev_q = q[t]
+
+    bursts = np.array(
+        [np.repeat(np.nan, N), np.repeat(offsets[0], N), np.repeat(offsets[0], N)],
+        ndmin=2,
+        dtype=object,
+    ).transpose()
+
+    burst_counter = -1
+    prev_q = 0
+    stack = np.zeros(int(N), dtype=int)
+    stack_counter = -1
+    for t in range(np.size(gaps)):
+        if q[t] > prev_q:
+            num_levels_opened = q[t] - prev_q
+            for i in range(int(num_levels_opened)):
+                burst_counter += 1
+                bursts[burst_counter, 0] = prev_q + i
+                bursts[burst_counter, 1] = offsets[t]
+                stack_counter += 1
+                stack[stack_counter] = int(burst_counter)
+        elif q[t] < prev_q:
+            num_levels_closed = prev_q - q[t]
+            for i in range(int(num_levels_closed)):
+                bursts[stack[stack_counter], 2] = offsets[t]
+                stack_counter -= 1
+        prev_q = q[t]
+
+    while stack_counter >= 0:
+        bursts[stack[stack_counter], 2] = offsets[np.size(gaps)]
+        stack_counter -= 1
+
+    return bursts
+
+
+def smooth_boolean_array(a: np.array, scale: int = 1) -> np.array:
     """
 
     Returns a boolean array in which isolated appearances of a feature are smoothed.
 
     Args:
         a (numpy.ndarray): Boolean instances.
+        scale (int): Kleinberg scale parameter. Higher values result in stricter smoothing.
 
     Returns:
         a (numpy.ndarray): Smoothened boolean instances.
 
     """
 
-    for i in range(1, len(a) - 1):
-        if a[i - 1] == a[i + 1]:
-            a[i] = a[i - 1]
-    return a == 1
+    offsets = np.where(a)[0]
+    if len(offsets) == 0:
+        return a  # no detected activity
+
+    bursts = kleinberg(offsets, gamma=0.01)
+    a = np.zeros(np.size(a), dtype=bool)
+    for i in bursts:
+        if i[0] == scale:
+            a[int(i[1]) : int(i[2])] = True
+
+    return a
 
 
 def split_with_breakpoints(a: np.ndarray, breakpoints: list) -> np.ndarray:
@@ -366,20 +554,22 @@ def split_with_breakpoints(a: np.ndarray, breakpoints: list) -> np.ndarray:
         breakpoints (list): list of breakpoints obtained with ruptures
 
     Returns:
-        split_a (np.ndarray): N (instances) * l (maximum break length) * m (features) shape
+        split_a (np.ndarray): padded array of shape N (instances) * l (maximum break length) * m (features)
 
     """
     rpt_lengths = list(np.array(breakpoints)[1:] - np.array(breakpoints)[:-1])
-    max_rpt_length = np.max([breakpoints[0], np.max(rpt_lengths)])
+
+    try:
+        max_rpt_length = np.max([breakpoints[0], np.max(rpt_lengths)])
+    except ValueError:
+        max_rpt_length = breakpoints[0]
 
     # Reshape experiment data according to extracted ruptures
     split_a = np.split(np.expand_dims(a, axis=0), breakpoints[:-1], axis=1)
 
     split_a = [
         np.pad(
-            i,
-            ((0, 0), (0, max_rpt_length - i.shape[1]), (0, 0)),
-            constant_values=0.0,
+            i, ((0, 0), (0, max_rpt_length - i.shape[1]), (0, 0)), constant_values=0.0
         )
         for i in split_a
     ]
@@ -389,7 +579,11 @@ def split_with_breakpoints(a: np.ndarray, breakpoints: list) -> np.ndarray:
 
 
 def rolling_window(
-    a: np.array, window_size: int, window_step: int, automatic_changepoints: str = False
+    a: np.array,
+    window_size: int,
+    window_step: int,
+    automatic_changepoints: str = False,
+    precomputed_breaks: np.ndarray = None,
 ) -> np.ndarray:
     """
 
@@ -401,6 +595,8 @@ def rolling_window(
         window_step (int): Step of the window to apply
         automatic_changepoints (str): Changepoint detection algorithm to apply.
         If False, applies a fixed sliding window.
+        precomputed_breaks (np.ndarray): Precomputed breaks to use, bypassing the changepoint detection algorithm.
+        None by default (break points are computed).
 
     Returns:
         rolled_a (np.ndarray): N (sliding window instances) * l (sliding window size) * m (features)
@@ -412,12 +608,17 @@ def rolling_window(
     if automatic_changepoints:
         # Define change point detection model using ruptures
         # Remove dimensions with low variance (occurring when aligning the animals with the y axis)
-        rpt_model = rpt.KernelCPD(
-            kernel=automatic_changepoints, min_size=window_size, jump=window_step
-        ).fit(VarianceThreshold(threshold=1e-3).fit_transform(a))
+        if precomputed_breaks is None:
+            rpt_model = rpt.KernelCPD(
+                kernel=automatic_changepoints, min_size=window_size, jump=window_step
+            ).fit(VarianceThreshold(threshold=1e-3).fit_transform(a))
+
+            # Extract change points from current experiment
+            breakpoints = rpt_model.predict(pen=4.0)
+
+        else:
+            breakpoints = np.cumsum(precomputed_breaks)
 
-        # Extract change points from current experiment
-        breakpoints = rpt_model.predict(pen=4.0)
         rolled_a = split_with_breakpoints(a, breakpoints)
 
     else:
@@ -437,6 +638,7 @@ def rupture_per_experiment(
     automatic_changepoints: str,
     window_size: int,
     window_step: int,
+    precomputed_breaks: dict = None,
 ) -> np.ndarray:
     """
 
@@ -456,6 +658,7 @@ def rupture_per_experiment(
         If not, it determines the minimum size of the obtained time series breaks.
         window_step (int): If automatic_changepoints is False, specifies the stride of the sliding window.
         If not, it determines the minimum step size of the obtained time series breaks.
+        precomputed_breaks (dict): If provided, changepoint detection is prevented, and provided breaks are used instead.
 
     Returns:
         ruptured_dataset (np.ndarray): Dataset with all ruptures concatenated across the first axis.
@@ -467,14 +670,19 @@ def rupture_per_experiment(
     ruptured_dataset, break_indices = None, None
     cumulative_shape = 0
     # Iterate over all experiments and populate them
-    for i, tab in enumerate(table_dict.values()):
+    for i, (key, tab) in enumerate(table_dict.items()):
         if i in rupture_indices:
             current_size = tab.shape[0]
             current_train, current_breaks = rolling_window(
-                to_rupture[cumulative_shape : cumulative_shape + current_size],
+                (
+                    to_rupture[cumulative_shape : cumulative_shape + current_size]
+                    if automatic_changepoints
+                    else to_rupture
+                ),
                 window_size,
                 window_step,
                 automatic_changepoints,
+                (None if not precomputed_breaks else precomputed_breaks[key]),
             )
             # Add shape of the current tab as the last breakpoint,
             # to avoid skipping breakpoints between experiments
@@ -506,7 +714,8 @@ def rupture_per_experiment(
                     )
 
                 # Once that's taken care of, concatenate ruptures alongside axis 0
-                ruptured_dataset = np.concatenate([ruptured_dataset, current_train])
+                if automatic_changepoints:
+                    ruptured_dataset = np.concatenate([ruptured_dataset, current_train])
                 if current_breaks is not None:
                     break_indices = np.concatenate([break_indices, current_breaks])
             except (ValueError, AttributeError):
@@ -653,7 +862,6 @@ def full_outlier_mask(
 
     for bpart in body_parts:
         if bpart != exclude:
-
             mask = mask_outliers(
                 experiment[bpart],
                 likelihood[bpart],
@@ -732,35 +940,35 @@ def filter_columns(columns: list, selected_id: str) -> list:
 
     """
 
+    if selected_id is None:
+        return columns
+
     columns_to_keep = []
     for column in columns:
         # Speed transformed columns
         if type(column) == str and column.startswith(selected_id):
             columns_to_keep.append(column)
         # Raw coordinate columns
-        if column[0].startswith(selected_id) and column[1] in [
-            "x",
-            "y",
-            "rho",
-            "phi",
-        ]:
+        if column[0].startswith(selected_id) and column[1] in ["x", "y", "rho", "phi"]:
             columns_to_keep.append(column)
         # Raw distance and angle columns
         elif len(column) in [2, 3] and all([i.startswith(selected_id) for i in column]):
             columns_to_keep.append(column)
+        elif column[0].lower().startswith("pheno"):
+            columns_to_keep.append(column)
 
     return columns_to_keep
 
 
 # noinspection PyUnboundLocalVariable
-def recognize_arena(
+def automatically_recognize_arena(
     videos: list,
     vid_index: int,
     path: str = ".",
     tables: dict = None,
     recoglimit: int = 500,
     high_fidelity: bool = False,
-    arena_type: str = "circular",
+    arena_type: str = "circular-autodetect",
     detection_mode: str = "rule-based",
     cnn_model: tf.keras.models.Model = None,
 ) -> Tuple[np.array, int, int]:
@@ -777,7 +985,7 @@ def recognize_arena(
         recoglimit (int): Number of frames to use for position estimates.
         high_fidelity (bool): If True, runs arena recognition on the whole video. Slow, but
         potentially more accurate in poor lighting conditions.
-        arena_type (string): Arena type; must be one of ['circular'].
+        arena_type (string): Arena type; must be one of ['circular-autodetect', 'circular-manual', 'polygon-manual'].
         detection_mode (str): Algorithm to use to detect the arena. "cnn" uses a
         pretrained model based on ResNet50 to predict the ellipse parameters from
         the image. "rule-based" (default) uses a simpler (and faster) image segmentation approach.
@@ -785,7 +993,9 @@ def recognize_arena(
 
     Returns:
         arena (np.ndarray): 1D-array containing information about the arena.
-        "circular" (3-element-array) -> x-y position of the center and the radius.
+        "circular-autodetect" (3-element-array) -> x-y position of the center and the radius.
+        "circular-manual" (3-element-array) -> x-y position of the center and the radius.
+        "polygon-manual" (2n-element-array) -> x-y position of each of the n the vertices of the polygon.
         h (int): Height of the video in pixels.
         w (int): Width of the video in pixels.
 
@@ -820,7 +1030,7 @@ def recognize_arena(
             print("Can't receive frame (stream end?). Exiting ...")
             break
 
-        if arena_type == "circular":
+        if arena_type == "circular-autodetect":
 
             # Detect arena and extract positions
             temp_center, temp_axes, temp_angle = circular_arena_recognition(
@@ -864,8 +1074,160 @@ def recognize_arena(
     return arena, h, w
 
 
+def retrieve_corners_from_image(frame: np.ndarray, arena_type: str):  # pragma: no cover
+    """
+
+    Opens a window and waits for the user to click on all corners of the polygonal arena.
+    The user should click on the corners in sequential order.
+
+    Args:
+        frame (np.ndarray): Frame to display.
+        arena_type (str): Type of arena to be used. Must be one of the following: "circular-manual", "polygon-manual".
+
+    Returns:
+
+        corners (np.ndarray): nx2 array containing the x-y coordinates of all n corners.
+
+    """
+
+    corners = []
+
+    def click_on_corners(event, x, y, flags, param):
+        # Callback function to store the coordinates of the clicked points
+        nonlocal corners, frame
+
+        if event == cv2.EVENT_LBUTTONDOWN:
+            corners.append((x, y))
+
+    # Create a window and display the image
+    cv2.startWindowThread()
+
+    while True:
+        frame_copy = frame.copy()
+
+        cv2.imshow(
+            "deepof - Select polygonal arena corners - (q: exit / d: delete)",
+            frame_copy,
+        )
+        cv2.setMouseCallback(
+            "deepof - Select polygonal arena corners - (q: exit / d: delete)",
+            click_on_corners,
+        )
+
+        # Display already selected corners
+        if len(corners) > 0:
+            for c, corner in enumerate(corners):
+                cv2.circle(frame_copy, (corner[0], corner[1]), 4, (40, 86, 236), -1)
+                # Display lines between the corners
+                if len(corners) > 1 and c > 0:
+                    if arena_type == "polygonal-manual" or len(corners) < 5:
+                        cv2.line(
+                            frame_copy,
+                            (corners[c - 1][0], corners[c - 1][1]),
+                            (corners[c][0], corners[c][1]),
+                            (40, 86, 236),
+                            2,
+                        )
+
+        # Close the polygon
+        if len(corners) > 2:
+            if arena_type == "polygonal-manual" or len(corners) < 5:
+                cv2.line(
+                    frame_copy,
+                    (corners[0][0], corners[0][1]),
+                    (corners[-1][0], corners[-1][1]),
+                    (40, 86, 236),
+                    2,
+                )
+        if len(corners) >= 5 and arena_type == "circular-manual":
+            cv2.ellipse(
+                frame_copy,
+                *fit_ellipse_to_polygon(corners),
+                startAngle=0,
+                endAngle=360,
+                color=(40, 86, 236),
+                thickness=3,
+            )
+
+        cv2.imshow(
+            "deepof - Select polygonal arena corners - (q: exit / d: delete)",
+            frame_copy,
+        )
+
+        # Remove last added coordinate if user presses 'd'
+        if cv2.waitKey(1) & 0xFF == ord("d"):
+            corners = corners[:-1]
+
+        # Exit is user presses q
+        elif cv2.waitKey(1) & 0xFF == ord("q"):
+            cv2.destroyAllWindows()
+            cv2.waitKey(1)
+            break
+
+    cv2.destroyAllWindows()
+    cv2.waitKey(1)
+
+    # Return the corners
+    return corners
+
+
+def extract_polygonal_arena_coordinates(
+    video_path: str, arena_type: str
+):  # pragma: no cover
+    """
+
+    Reads a random frame from the selected video, and opens an interactive GUI to let the user delineate
+    the arena manually.
+
+    Args:
+        video_path: Path to the video file.
+        arena_type: Type of arena to be used. Must be one of the following: "circular-manual", "polygon-manual".
+
+    Returns:
+        np.ndarray: nx2 array containing the x-y coordinates of all n corners of the polygonal arena.
+        int: Height of the video.
+        int: Width of the video.
+
+    """
+
+    current_video = imread(video_path)
+    current_frame = np.random.choice(current_video.shape[0])
+
+    # Get and return the corners of the arena
+    arena_corners = retrieve_corners_from_image(
+        current_video[current_frame].compute(), arena_type
+    )
+    return arena_corners, current_video.shape[2], current_video.shape[1]
+
+
+def fit_ellipse_to_polygon(polygon: list):  # pragma: no cover
+    """
+
+    Fits an ellipse to the provided polygon.
+
+    Args:
+        polygon: List of (x,y) coordinates of the corners of the polygon.
+
+    Returns:
+        tuple: (x,y) coordinates of the center of the ellipse.
+        tuple: (a,b) semi-major and semi-minor axes of the ellipse.
+        float: Angle of the ellipse.
+
+    """
+
+    # Detect the main ellipse containing the arena
+    ellipse_params = cv2.fitEllipse(np.array(polygon))
+
+    # Parameters to return
+    center_coordinates = tuple([int(i) for i in ellipse_params[0]])
+    axes_length = tuple([int(i) // 2 for i in ellipse_params[1]])
+    ellipse_angle = ellipse_params[2]
+
+    return center_coordinates, axes_length, ellipse_angle
+
+
 def circular_arena_recognition(
-    frame: np.array,
+    frame: np.ndarray,
     detection_mode: str = "rule-based",
     cnn_model: tf.keras.models.Model = None,
 ) -> np.array:
@@ -902,13 +1264,9 @@ def circular_arena_recognition(
         )
         main_cnt = np.argmax([len(c) for c in cnts])
 
-        # Detect the main ellipse containing the arena
-        ellipse_params = cv2.fitEllipse(cnts[main_cnt])
-
-        # Parameters to return
-        center_coordinates = tuple([int(i) for i in ellipse_params[0]])
-        axes_length = tuple([int(i) // 2 for i in ellipse_params[1]])
-        ellipse_angle = ellipse_params[2]
+        center_coordinates, axes_length, ellipse_angle = fit_ellipse_to_polygon(
+            cnts[main_cnt]
+        )
 
     elif detection_mode == "cnn":
 
@@ -968,7 +1326,7 @@ def rolling_speed(
 
     original_shape = dframe.shape
     if center:
-        body_parts = [bp for bp in dframe.columns.levels[0] if bp != center]
+        body_parts = [bp for bp in dframe.columns.levels[0] if center not in bp]
     else:
         try:
             body_parts = dframe.columns.levels[0]
@@ -1157,7 +1515,3 @@ def cluster_transition_matrix(
         return trans_normed, autocorr
 
     return trans_normed
-
-
-# TODO:
-#    - Add center / time in zone to supervised_tagging
